from __future__ import absolute_import

import os
import uuid
from datetime import datetime

from restart.exceptions import BadRequest, MethodNotAllowed
from restart.ext.mongo.collection import Collection
from restart.ext.mongo.utils import ensure_dir


class Files(Collection):
    """The resource class for managing uploaded files.

    Storage scheme:
        1. The file content will be stored to server's file system.
        2. The file meta-data and the (extra) form data will be saved
           into the database.

        The document schema:
            {
                initial_name (The initial file name)
                storage_path (The final storage path)
                date_uploaded (The date uploaded)
                ... (The form data from request)
            }
    """

    #: The folder for storing the uploaded files.
    #: Must be a string.
    upload_folder = None

    #: The file extensions allowed to upload.
    #: Must be a tuple of strings.
    allowed_extensions = None

    #: The string format of the archive name.
    #: An archive is a folder that includes files, which are
    #: uploaded in the same time period.
    #:
    #: For example:
    #:
    #:     '%Y%m': in the same month
    #:     '%Y%m%d': on the same day (the default value)
    #:     '%Y%m%d%H': within the same hour
    archive_name_format = '%Y%m%d'

    def allowed_file(self, filename):
        assert isinstance(self.allowed_extensions, tuple), \
            'The `allowed_extensions` attribute must be a tuple object'
        return ('.' in filename and
                filename.rsplit('.', 1)[1] in self.allowed_extensions)

    def get_storage_path(self, filename, date_uploaded):
        uid = uuid.uuid4()
        _, ext = os.path.splitext(filename)
        filename = '%s%s' % (uid, ext)

        archive_name = date_uploaded.strftime(self.archive_name_format)
        storage_path = os.path.join(archive_name, filename)

        return storage_path

    def get_full_storage_path(self, storage_path):
        assert isinstance(self.upload_folder, basestring), \
            'The `upload_folder` attribute must be a string'
        return os.path.join(self.upload_folder, storage_path)

    def save_file(self, fileobj, storage_path):
        full_storage_path = self.get_full_storage_path(storage_path)
        ensure_dir(os.path.dirname(full_storage_path))

        with open(full_storage_path, 'wb') as f:
            f.write(fileobj.read())

    def remove_file(self, storage_path):
        full_storage_path = self.get_full_storage_path(storage_path)
        if os.path.isfile(full_storage_path):
            os.remove(full_storage_path)

    def get_fileobj(self, files):
        if len(files) != 1:
            raise BadRequest('One and only one file is allowed')

        fileobj = files.values()[0]
        if not self.allowed_file(fileobj.filename):
            raise BadRequest('Only file extensions (%s) are allowed' %
                             ', '.join(self.allowed_extensions))

        return fileobj

    def get_file_metadata(self, fileobj):
        """Get the file meta-data from the file object."""
        date_uploaded = datetime.now()
        storage_path = self.get_storage_path(fileobj.filename, date_uploaded)

        return dict(
            initial_name=fileobj.filename,
            storage_path=storage_path,
            date_uploaded=date_uploaded
        )

    def update_request(self, request, file_metadata):
        """Update the request object.

        Now only merge the file meta-data into the request data.
        Override to implement your own strategy.
        """
        request.data.update(file_metadata)
        return request

    def create(self, request):
        """Create a new file."""
        fileobj = self.get_fileobj(request.files)

        # Store the file content to the file system
        file_metadata = self.get_file_metadata(fileobj)
        self.save_file(fileobj, file_metadata['storage_path'])

        # Save the file meta-data and the form data into the database
        request = self.update_request(request, file_metadata)
        return super(Files, self).create(request)

    def replace(self, request, pk):
        """Replace the old file with a new one.

        You can set the `X-Keep-Storage-Path` headr as follows:

            X-Keep-Storage-Path: true

        to replace the file while still using the old storage path.
        """
        fileobj = self.get_fileobj(request.files)
        doc = self.get_doc(pk)

        # Remove the old file
        self.remove_file(doc['storage_path'])

        # Use the old storage path if `X-Keep-Storage-Path` headr is set
        file_metadata = self.get_file_metadata(fileobj)
        if request.headers.get('X-Keep-Storage-Path'):
            file_metadata['storage_path'] = doc['storage_path']

        # Store the new file content to the file system
        self.save_file(fileobj, file_metadata['storage_path'])

        # Save the new file meta-data and the form data into the database
        request = self.update_request(request, file_metadata)
        return super(Files, self).replace(request, pk)

    def update(self, request, pk):
        raise MethodNotAllowed()

    def delete(self, request, pk):
        """Delete the file."""
        doc = self.get_doc(pk)
        self.remove_file(doc['storage_path'])
        return super(Files, self).delete(request, pk)



from __future__ import absolute_import

import copy

import bson
from jsonpatch import JsonPatch
from restart import status
from restart.resource import Resource
from restart.exceptions import NotFound
from restart.utils import make_location_header

from .form import Form
from .filter import Filter
from .paginator import Paginator
from .sorter import Sorter
from .selector import Selector
from .middlewares import SerializerMiddleware
from .utils import get_exception_detail


class Collection(Resource):
    """The resource class that represents a MongoDB collection.

    Note:
       You must subclass this class, and specify the
       class-level attributes `database` and `collection_name`.
    """

    #: The database instance
    database = None

    #: The collection name
    collection_name = None

    #: The validation form class
    form_class = Form

    #: The filter class
    filter_class = Filter

    #: The paginator object
    paginator = Paginator()

    #: The sorter object
    sorter = Sorter()

    #: The selector object
    selector = Selector()

    middleware_classes = (SerializerMiddleware,)

    def __init__(self, *args, **kwargs):
        if self.database is None:
            raise AttributeError('The class-level attribute `database` '
                                 'must be specified.')
        if self.collection_name is None:
            raise AttributeError('The class-level attribute `collection_name` '
                                 'must be specified.')

        self.engine = self.database[self.collection_name]
        self.filter = self.filter_class(self.database, self.collection_name)

        super(Collection, self).__init__(*args, **kwargs)

    def get_pk(self, pk):
        try:
            return bson.ObjectId(pk)
        except bson.errors.InvalidId:
            raise NotFound()

    def get_doc(self, pk, fields=None):
        doc = self.engine.find_one(
            {'_id': self.get_pk(pk)},
            fields=fields
        )
        if doc:
            return doc
        else:
            raise NotFound()

    def get_params(self, request):
        query_params = copy.deepcopy(request.args)
        page, per_page = self.paginator.get_args(query_params)
        sort = self.sorter.get_args(query_params)
        fields = self.selector.get_fields(query_params)
        lookup = self.filter.query(query_params)
        return page, per_page, sort, fields, lookup

    def index(self, request):
        page, per_page, sort, fields, lookup = self.get_params(request)

        skip, limit = (page - 1) * per_page, per_page
        docs = self.engine.find(
            spec=lookup, skip=skip, limit=limit,
            sort=sort, fields=fields
        )
        count = self.engine.find(spec=lookup).count()
        headers = self.paginator.make_headers(
            request.uri.split('?', 1)[0],
            page, per_page, count
        )
        return list(docs), status.HTTP_200_OK, headers

    def create(self, request):
        form = self.form_class(request.data)
        if form.is_valid():
            _id = self.engine.insert(form.document)
            headers = {'Location': make_location_header(request, _id)}
            return {'_id': _id}, status.HTTP_201_CREATED, headers
        return form.errors, status.HTTP_400_BAD_REQUEST

    def read(self, request, pk):
        _, _, _, fields, _ = self.get_params(request)
        return self.get_doc(pk, fields)

    def replace(self, request, pk):
        form = self.form_class(request.data)
        if form.is_valid():
            doc = form.document
            doc.update({'_id': self.get_pk(pk)})
            self.engine.remove({'_id': doc['_id']})
            self.engine.insert(doc)
            return '', status.HTTP_204_NO_CONTENT
        return form.errors, status.HTTP_400_BAD_REQUEST

    def update(self, request, pk):
        doc = self.get_doc(pk)

        # Do JSON-Patch
        patch_data = JsonPatch(request.data)
        try:
            patch_data.apply(doc, in_place=True)
        except Exception as e:
            return ({'jsonpatch_error': get_exception_detail(e)},
                    status.HTTP_400_BAD_REQUEST)

        # Validate data after JSON-Patch
        form = self.form_class(doc)
        if form.is_valid():
            self.engine.save(form.document)
            return '', status.HTTP_204_NO_CONTENT
        return form.errors, status.HTTP_400_BAD_REQUEST

    def delete(self, request, pk):
        doc = self.get_doc(pk)
        self.engine.remove({'_id': doc['_id']})
        return '', status.HTTP_204_NO_CONTENT



from __future__ import absolute_import

import functools


def query_params(*names):
    """A helper decorator for binding query parameters specified by
    `names` as arguments to `query_<name>` methods.

    For example:

        class UserFilter(Filter):
            @query_params('username')
            def query_username(self, username):
                return {'username': username.upper()}

            @query_params('date_joined_gt', 'date_joined_lt')
            def query_datejoined(self, date_joined_gt=None, date_joined_lt=None):
                conditions = {}
                if date_joined_gt:
                    conditions.update({'$gt': date_joined_gt})
                if date_joined_lt:
                    conditions.update({'$lt': date_joined_lt})
                if conditions:
                    return {'date_joined': conditions}
                else:
                    return {}
    """
    def wrapper(method):
        @functools.wraps(method)
        def decorator(self, params):
            kwargs = {}
            for name in names:
                if name in params:
                    value = params.pop(name)
                    kwargs.update({name: value})
            try:
                return method(self, **kwargs)
            except TypeError:
                return {}
        return decorator

    return wrapper


class Filter(object):
    """The Class used for customized filtering.

    1. Raw query parameter

    Raw query parameter supports simple filtering:

        /users?name=russellluo

    Note that the MongoDB-style nested field names are also supported):

        /users?name=russellluo&address.country=China

    2. The `Filter` class (and its subclasses)

    The `Filter` class (and its subclasses) supports user-defined methods for customized filtering:

        1. override `query` method
        2. add `query_<name>` method

        All methods should return MongoDB-style conditions.

    For example,
    """

    def __init__(self, database, collection_name):
        self.database = database
        self.collection_name = collection_name

    def merge(self, conditions):
        """Merge `conditions` using AND-logic into one condition."""
        if conditions:
            return {'$and': conditions}
        else:
            return {}

    def query(self, params):
        """Generate lookup conditions."""
        conditions = []

        # Collect conditions from `query_<name>` methods
        for attr_name in dir(self):
            if attr_name.startswith('query_'):
                method = getattr(self, attr_name)
                condition = method(params)
                if condition:
                    conditions.append(condition)

        # Also treat remaining `params` as conditions
        if params:
            conditions.append(params)

        return self.merge(conditions)



from __future__ import absolute_import

from restart.ext.mongo.filter import Filter


class DuFilter(Filter):
    """The class for double-underscore style filtering."""

    separator = '__'

    def get_du_params(self, params, suffix):
        """Get double-underscore style parameters which match `suffix`.

        For example:
            params: {'age__lt': 20, 'age__gt': 10, 'level__lt': 5}
            suffix: 'lt' => du_params: {'age': 20, 'level': 5}
            suffix: 'gt' => du_params: {'age': 10}
        """
        keys = [
            key for key in params
            if key.endswith(self.separator + suffix)
        ]
        du_params = {
            key.split(self.separator)[0]: params.pop(key)
            for key in keys
        }
        return du_params

    def get_conditions(self, params, suffix, comparer, convertor=None):
        if convertor is None:
            convertor = lambda v: v

        du_params = self.get_du_params(params, suffix)
        conditions = {
            key: {comparer: convertor(value)}
            for key, value in du_params.items()
        }
        return conditions

    def query_ne(self, params):
        """Query by keyword `<fieldname>__ne`.

        For example:
            /?age__ne=int(20) <=> {'age': {'$ne': 20}}
        """
        return self.get_conditions(params, 'ne', '$ne')

    def query_lt(self, params):
        """Query by keyword `<fieldname>__lt`.

        For example:
            /?age__lt=int(20) <=> {'age': {'$lt': 20}}
        """
        return self.get_conditions(params, 'lt', '$lt')

    def query_lte(self, params):
        """Query by keyword `<fieldname>__lte`.

        For example:
            /?age__lte=int(20) <=> {'age': {'$lte': 20}}
        """
        return self.get_conditions(params, 'lte', '$lte')

    def query_gt(self, params):
        """Query by keyword `<fieldname>__gt`.

        For example:
            /?age__gt=int(20) <=> {'age': {'$gt': 20}}
        """
        return self.get_conditions(params, 'gt', '$gt')

    def query_gte(self, params):
        """Query by keyword `<fieldname>__gte`.

        For example:
            /?age__gte=int(20) <=> {'age': {'$gte': 20}}
        """
        return self.get_conditions(params, 'gte', '$gte')

    def query_in(self, params):
        """Query by keyword `<fieldname>__in`.

        For example:
            /?age__in=int(20) <=> {'age': {'$in': [20]}}
            /?age__in=int(20)&age__in=int(25) <=> {'age': {'$in': [20, 25]}}
        """
        def make_list(value):
            if not isinstance(value, (tuple, list)):
                return [value]
            return value

        return self.get_conditions(params, 'in', '$in', make_list)

    def query_like(self, params):
        """Query by keyword `<fieldname>__like`.

        For example:
            /?name__like=^russ <=> {'name': {'$regex': '^russ'}}
        """
        return self.get_conditions(params, 'like', '$regex')



from __future__ import absolute_import

import csv
import codecs
from cStringIO import StringIO

from restart.renderers import Renderer


class UnicodeCSVWriter(object):
    """A CSV writer that can write Unicode rows to the CSV file,
    which is encoded in the given encoding.

    According to https://docs.python.org/2/library/csv.html::
        The standard version of the `csv` module does not
        support Unicode input.

    Borrowed from https://docs.python.org/2/library/csv.html#examples.
    """
    def __init__(self, csvfile, dialect=csv.excel,
                 encoding='utf-8', **kwargs):
        self.queue = StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwargs)
        self.stream = csvfile
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([unicode(s).encode('utf-8') for s in row])

        # Fetch utf-8 output from the queue
        data = self.queue.getvalue()
        data = data.decode('utf-8')
        # And reencode it into the target encoding
        data = self.encoder.encode(data)

        # Write to the target stream
        self.stream.write(data)

        # Empty the queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


class CSVRenderer(Renderer):
    """The CSV renderer class."""

    #: A tuple that specifies the columns in the CSV.
    #:
    #: Each item of `columns` is also a tuple in the form of
    #: `(header, fieldname)` or `(header, fieldname, converter)`.
    #:
    #: - header: The column header.
    #: - fieldname: The name of the field in the database, which
    #:              provides the original column value. To represent
    #:              a nested field, use dot (`.`) in its name.
    #: - converter: The converter used to make the final column value.
    #:              Besides the original `value`, an additional `context`,
    #:              which is a dictionary in the form of ``{
    #:                  'data': <The whole row data>,
    #:                  'renderer': <The renderer context>
    #:              }``, will be passed into the converter callable.
    #:
    #: The converter is determined in the following order:
    #:
    #:     1. The converter defined in the 3-tuple. Any function with
    #:        the following prototype can be specified as a converter
    #:        in the form of 3-tuple:
    #:
    #:         def converter(value, context):
    #:             return converted_value
    #:
    #:     2. The converter defined as a method, which has the following
    #:        prototype:
    #:
    #:         def convert_<fieldname>(self, value, context):
    #:             return converted_value
    #:
    #:        Note that the `fieldname` part is the column fieldname with
    #:        all its dots (`.`) transformed to underscores (`_`).
    #:
    #:     3. If not specified, use the default converter.
    #:
    #: For example:
    #:
    #:     def capitalize(value, context):
    #:         return value.capitalize()
    #:
    #:     class UserCSVRenderer(CSVRenderer):
    #:
    #:         columns = (
    #:             # A simple fieldname with the function-style converter
    #:             ('name', 'username', capitalize),
    #:             # A simple fieldname with the default converter
    #:             ('age', 'age'),
    #:             # A nested fieldname with the method-style converter
    #:             ('phone', 'contact.phone'),
    #:             ...
    #:         )
    #:
    #:         def convert_contact_phone(self, value, context):
    #:             return unicode('086-%s' % value)
    columns = None

    #: The default value for the missing field specified in the `columns`
    default_value = ''

    #: The encoding of the final CSV
    encoding = 'utf-8'

    #: The content type bound to this renderer.
    content_type = 'text/csv'

    #: The format suffix bound to this renderer.
    format_suffix = 'csv'

    def default_converter(self, value, context):
        """The default converter."""
        return unicode(value)

    def get_schema(self):
        """Get the CSV schema."""
        assert isinstance(self.columns, tuple), \
            'The `columns` attribute must be a tuple object'

        headers = []
        fields = []

        for column in self.columns:
            if len(column) == 2:
                header, fieldname = column
                methodname = 'convert_%s' % fieldname.replace('.', '_')
                converter = getattr(self, methodname, self.default_converter)
            else:
                header, fieldname, converter = column

            headers.append(header)

            keys = fieldname.split('.')
            fields.append((keys, converter))

        return headers, fields

    def render(self, data, context=None):
        """Render `data` into CSV.

        :param data: the data to be rendered.
        :param context: a dictionary containing extra context data
                        that can be useful for rendering.
        """
        assert isinstance(data, (dict, list, tuple)), \
            'The `data` argument must be a dict or a list or a tuple'

        # Assure that data is a sequence
        if isinstance(data, dict):
            data = [data]

        headers, fields = self.get_schema()

        csv_file = StringIO()
        csv_writer = UnicodeCSVWriter(csv_file, encoding=self.encoding)

        # Write the headers
        csv_writer.writerow(headers)

        # Write the rows
        for each in data:
            values = []

            # Extract the values
            for keys, converter in fields:
                # Get the value of `fieldname` from `each`
                value = each
                for key in keys:
                    value = value.get(key)
                    if value is None:
                        value = self.default_value
                        break

                # Convert the value
                converter_context = {'data': each, 'renderer': context}
                value = converter(value, converter_context)

                values.append(value)

            # Write the values
            csv_writer.writerow(values)

        csv_data = csv_file.getvalue()
        return csv_data



from __future__ import absolute_import


class Selector(object):
    """The class for selecting which fields are returned.

    This class supports the keyword `fields` in query parameter.

    For example, the following request will receive all users with only
    `name` and `password` fields in each user:

        /users?fields=name,password

    Note that the MongoDB-style nested field names are also supported:

        /users?fields=name,address.country
    """

    def get_raw(self, query_params):
        """Get the raw fields specified in the query parameters."""
        fields = query_params.pop('fields', None)
        if fields is None:
            raw = None
        else:
            raw = fields.split(',')
        return raw

    def get_fields(self, query_params):
        """Get the MongoDB-style fields from the query parameters."""
        raw = self.get_raw(query_params)
        if raw is None:
            return None

        fields = {
            field_name: True
            for field_name in raw
        }
        if raw and '_id' not in raw:
            fields.update({'_id': False})
        return fields



from __future__ import absolute_import

from jsonsir import Serializer
from jsonsir.contrib.intencoder import IntEncoder
from jsonsir.contrib.boolencoder import BoolEncoder
from jsonsir.contrib.regexencoder import RegexEncoder
from jsonsir.contrib.objectidencoder import ObjectIdEncoder
from jsonsir.contrib.datetimeencoder import DateTimeEncoder


# Make an instance of `Serializer` (bound with specified encoders)
serializer = Serializer([
    IntEncoder(),
    BoolEncoder(),
    RegexEncoder(),
    ObjectIdEncoder(),
    # Using UTC Format, see http://www.w3.org/TR/NOTE-datetime for details
    DateTimeEncoder('%Y-%m-%dT%H:%M:%SZ'),
])


class SerializerMiddleware(object):
    """The middleware used for serializing/deserializing the
    request/response payload.
    """

    #: The serializer object
    serializer = serializer

    #: The serializing scheme
    #:     True -- serialize data and bind it with type names
    #:     False -- only serialize data
    with_type_name = False

    def process_request(self, request):
        """Deserialize the request data and query parameters.

        :param request: the request object.
        """
        request._data = self.serializer.deserialize(request.data)
        request._args = self.serializer.deserialize(request.args)

    def process_response(self, request, response):
        """Serialize the response data.

        :param request: the request object.
        :param response: the response object.
        """
        response.data = self.serializer.serialize(response.data,
                                                  self.with_type_name)
        return response



from __future__ import absolute_import


class Sorter(object):
    """The class for sorting.

    This class supports the keyword `sort` in query parameter:

    Sign       | Order
    ---------- | ----------
    + or empty | Ascending
    -          | Descending

    For example, you can sort users first by `name` in ascending-order
    and second by `age` in descending-order::

        /users?sort=name,-age

    Note that the MongoDB-style nested field names are also supported:

        /users?sort=name,address.country
    """

    def get_args(self, query_params):
        sort = query_params.pop('sort', None)
        if sort is None:
            args = None
        else:
            args = [
                (arg[1:], -1) if arg.startswith('-') else (arg, 1)
                for arg in sort.split(',')
            ]
        return args



from __future__ import absolute_import


class Paginator(object):
    """The class for pagination.

    This class supports the following two keywords in query parameter:

    Keyword  | Default
    -------- | -------
    page     | 1
    per_page | 10

    You may want to specify `page` and `per_page` explicitly:

        /users?page=2&per_page=10
    """

    #: The number of items per page
    per_page = 10

    def get_args(self, query_params):
        PAGE = 1
        PER_PAGE = self.per_page

        try:
            page = int(query_params.pop('page', None))
            if page < 1:
                page = PAGE
        except Exception:
            page = PAGE

        try:
            per_page = int(query_params.pop('per_page', None))
            if per_page < 1:
                per_page = PER_PAGE
        except Exception:
            per_page = PER_PAGE

        return page, per_page

    def make_headers(self, uri, page, per_page, count):
        """Make GitHub-style pagination headers."""
        headers = {}

        div, mod = divmod(count, per_page)
        page_count = div + int(mod > 0)

        links = []

        def add_link(page, per_page, rel):
            args = (uri, page, per_page, rel)
            links.append('<%s?page=%s&per_page=%s>; rel="%s"' % args)

        def add_prev_link():
            add_link(page - 1, per_page, 'prev')

        def add_next_link():
            add_link(page + 1, per_page, 'next')

        def add_first_link():
            add_link(1, per_page, 'first')

        def add_last_link():
            add_link(page_count, per_page, 'last')

        # fill `Link` headers
        if page_count > 0 and page != page_count:
            if page == 1:
                add_next_link()
                add_last_link()
            elif page < page_count:
                add_prev_link()
                add_next_link()
                add_first_link()
                add_last_link()
            elif page == page_count:
                add_prev_link()
                add_first_link()
            else:
                add_link(page_count, per_page, 'prev')
                add_first_link()
                add_last_link()

            headers.update({'Link': ', '.join(links)})

        # fill `X-Pagination-Info` headers
        args = (page, per_page, count)
        headers.update({
            'X-Pagination-Info': 'page=%s, per-page=%s, count=%s' % args
        })

        return headers



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages


def get_info(name):
    basedir = os.path.dirname(__file__)
    with open(os.path.join(basedir, 'restart_mongo/__init__.py')) as f:
        locals = {}
        exec(f.read(), locals)
        return locals[name]


setup(
    name='RESTArt-Mongo',
    version=get_info('__version__'),
    author=get_info('__author__'),
    author_email=get_info('__email__'),
    maintainer=get_info('__author__'),
    maintainer_email=get_info('__email__'),
    keywords='REST APIs, Python',
    description=get_info('__doc__'),
    license=get_info('__license__'),
    long_description=get_info('__doc__'),
    packages=find_packages(exclude=['tests']),
    url='https://github.com/RussellLuo/restart-mongo',
    install_requires=[
        'python-restart',
        'pymongo==2.7.2',
        'jsonpatch==1.7',
        'JsonForm>=0.0.2',
        'JsonSir>=0.0.2',
    ],
)



from __future__ import absolute_import

import os
import errno


def get_exception_detail(e):
    if not isinstance(e, Exception):
        return u''
    return u'%s: %s' % (e.__class__.__name__, e.message)


def ensure_dir(dirname):
    """Ensure the directory exists.

    No error if existing, make parent directories as needed.
    """
    try:
        os.makedirs(dirname)
    except OSError, e:
        if e.errno != errno.EEXIST:
            raise e
